AVOGADRO = 6.02214076e23  # mol^-1
HARTREE2INVERSE_M = 2.194746313705E7  # m^-1
HARTREE2INVERSE_CM = 100 / HARTREE2INVERSE_M
ELECTRON_CHARGE = 1.602176487e-19  # C
ELECTRON_MASS = 9.10938215E-31  # kg
EPSILON_0 = 8.8541878128E-12  # F m^-1
SPEED_OF_LIGHT_M_S = 299792458.0  # m s^-1
SPEED_OF_LIGHT_CM_S = SPEED_OF_LIGHT_M_S * 100.0  # cm s^-1
EV_TO_NM = 1239.84193  # eV nm
EV_TO_WAVENUMBER = 8065.54429  # eV^-1 cm^-1
